@@ -1,7 +1,7 @@
 /*
  *  signing.c
  *
- *  Copyright (c) 2008-2018 Pacman Development Team <pacman-dev@archlinux.org>
+ *  Copyright (c) 2008-2019 Pacman Development Team <pacman-dev@archlinux.org>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -182,7 +182,7 @@ static int init_gpgme(alpm_handle_t *handle)
 #endif
 	/* NOTE:
 	 * The GPGME library installs a SIGPIPE signal handler automatically if
-	 * the default signal hander is in use. The only time we set a handler
+	 * the default signal handler is in use. The only time we set a handler
 	 * for SIGPIPE is in dload.c, and we reset it when we are done. Given that
 	 * we do this, we can let GPGME do its automagic. However, if we install
 	 * a library-wide SIGPIPE handler, we will have to be careful.
@@ -258,7 +258,46 @@ error:
 }
 
 /**
- * Search for a GPG key in a remote location.
+ * Import a key from a Web Key Directory (WKD) into the local keyring using.
+ * This requires GPGME to call the gpg binary.
+ * @param handle the context handle
+ * @param email the email address of the key to import
+ * @return 0 on success, -1 on error
+ */
+static int key_import_wkd(alpm_handle_t *handle, const char *email)
+{
+	gpgme_error_t gpg_err;
+	gpgme_ctx_t ctx;
+	gpgme_keylist_mode_t mode;
+	gpgme_key_t key;
+	int ret = -1;
+
+	memset(&ctx, 0, sizeof(ctx));
+	gpg_err = gpgme_new(&ctx);
+	CHECK_ERR();
+
+	mode = gpgme_get_keylist_mode(ctx);
+	mode |= GPGME_KEYLIST_MODE_LOCATE;
+	gpg_err = gpgme_set_keylist_mode(ctx, mode);
+	CHECK_ERR();
+
+	_alpm_log(handle, ALPM_LOG_DEBUG, _("looking up key %s using WKD\n"), email);
+	gpg_err = gpgme_get_key(ctx, email, &key, 0);
+	if(gpg_err_code(gpg_err) == GPG_ERR_NO_ERROR) {
+		ret = 0;
+	}
+	gpgme_key_unref(key);
+
+gpg_error:
+	if(ret != 0) {
+		_alpm_log(handle, ALPM_LOG_DEBUG, _("gpg error: %s\n"), gpgme_strerror(gpg_err));
+	}
+	gpgme_release(ctx);
+	return ret;
+}
+
+/**
+ * Search for a GPG key on a keyserver.
  * This requires GPGME to call the gpg binary and have a keyserver previously
  * defined in a gpg.conf configuration file.
  * @param handle the context handle
@@ -266,7 +305,7 @@ error:
  * @param pgpkey storage location for the given key if found
  * @return 1 on success, 0 on key not found, -1 on error
  */
-static int key_search(alpm_handle_t *handle, const char *fpr,
+static int key_search_keyserver(alpm_handle_t *handle, const char *fpr,
 		alpm_pgpkey_t *pgpkey)
 {
 	gpgme_error_t gpg_err;
@@ -386,10 +425,10 @@ gpg_error:
 /**
  * Import a key into the local keyring.
  * @param handle the context handle
- * @param key the key to import, likely retrieved from #key_search
+ * @param key the key to import, likely retrieved from #key_search_keyserver
  * @return 0 on success, -1 on error
  */
-static int key_import(alpm_handle_t *handle, alpm_pgpkey_t *key)
+static int key_import_keyserver(alpm_handle_t *handle, alpm_pgpkey_t *key)
 {
 	gpgme_error_t gpg_err;
 	gpgme_ctx_t ctx;
@@ -430,45 +469,85 @@ gpg_error:
 	return ret;
 }
 
+/** Extract the email address from a user ID
+ * @param uid the user ID to parse in the form "Example Name <email@address.invalid>"
+ * @param email to hold email address
+ * @return 0 on success, -1 on error
+ */
+static int email_from_uid(const char *uid, char **email)
+{
+       char *start, *end;
+
+       if (uid == NULL) {
+               email = NULL;
+               return -1;
+       }
+
+       start = strrchr(uid, '<');
+       if(start) {
+               end = strrchr(start, '>');
+       }
+
+       if(start && end) {
+               STRNDUP(*email, start+1, end-start-1, return -1);
+               return 0;
+       } else {
+               email = NULL;
+               return -1;
+       }
+}
+
 /**
  * Import a key defined by a fingerprint into the local keyring.
  * @param handle the context handle
+ * @param uid a user ID of the key to import
  * @param fpr the fingerprint key ID to import
  * @return 0 on success, -1 on error
  */
-int _alpm_key_import(alpm_handle_t *handle, const char *fpr)
+int _alpm_key_import(alpm_handle_t *handle, const char *uid, const char *fpr)
 {
 	int ret = -1;
 	alpm_pgpkey_t fetch_key;
-	memset(&fetch_key, 0, sizeof(fetch_key));
+	char *email;
 
-	if(key_search(handle, fpr, &fetch_key) == 1) {
-		_alpm_log(handle, ALPM_LOG_DEBUG,
-				"unknown key, found %s on keyserver\n", fetch_key.uid);
-		if(!_alpm_access(handle, handle->gpgdir, "pubring.gpg", W_OK)) {
-			alpm_question_import_key_t question = {
+	if(_alpm_access(handle, handle->gpgdir, "pubring.gpg", W_OK)) {
+		/* no chance of import succeeding if pubring isn't writable */
+		_alpm_log(handle, ALPM_LOG_ERROR, _("keyring is not writable\n"));
+		return -1;
+	}
+
+	memset(&fetch_key, 0, sizeof(fetch_key));
+	STRDUP(fetch_key.uid, uid, return -1);
+	STRDUP(fetch_key.fingerprint, fpr, return -1);
+
+	alpm_question_import_key_t question = {
 				.type = ALPM_QUESTION_IMPORT_KEY,
 				.import = 0,
 				.key = &fetch_key
 			};
-			QUESTION(handle, &question);
-			if(question.import) {
-				if(key_import(handle, &fetch_key) == 0) {
+	QUESTION(handle, &question);
+	if(question.import) {
+		/* Try to import the key from a WKD first */
+		if(email_from_uid(uid, &email) == 0) {
+			ret = key_import_wkd(handle, email);
+		}
+
+		/* If importing from the WKD fails, fall back to keyserver lookup */
+		if(ret != 0) {
+			if(key_search_keyserver(handle, fpr, &fetch_key) == 1) {
+				_alpm_log(handle, ALPM_LOG_DEBUG,
+						_("key \"%s\" on keyserver\n"), fetch_key.uid);
+				if(key_import_keyserver(handle, &fetch_key) == 0) {
 					ret = 0;
 				} else {
 					_alpm_log(handle, ALPM_LOG_ERROR,
 							_("key \"%s\" could not be imported\n"), fetch_key.uid);
 				}
+			} else {
+				_alpm_log(handle, ALPM_LOG_ERROR,
+						_("key \"%s\" could not be looked up remotely\n"), fpr);
 			}
-		} else {
-			/* keyring directory was not writable, so we don't even try */
-			_alpm_log(handle, ALPM_LOG_WARNING,
-					_("key %s, \"%s\" found on keyserver, keyring is not writable\n"),
-					fetch_key.fingerprint, fetch_key.uid);
 		}
-	} else {
-		_alpm_log(handle, ALPM_LOG_ERROR,
-				_("key \"%s\" could not be looked up remotely\n"), fpr);
 	}
 	gpgme_key_unref(fetch_key.data);
 
@@ -709,19 +788,24 @@ error:
 }
 
 #else /* HAVE_LIBGPGME */
-int _alpm_key_in_keychain(alpm_handle_t UNUSED *handle, const char UNUSED *fpr)
+int _alpm_key_in_keychain(alpm_handle_t *handle, const char UNUSED *fpr)
 {
+	handle->pm_errno = ALPM_ERR_MISSING_CAPABILITY_SIGNATURES;
 	return -1;
 }
 
-int _alpm_key_import(alpm_handle_t UNUSED *handle, const char UNUSED *fpr)
+int _alpm_key_import(alpm_handle_t *handle, const char UNUSED *uid,
+		const char UNUSED *fpr)
 {
+	handle->pm_errno = ALPM_ERR_MISSING_CAPABILITY_SIGNATURES;
 	return -1;
 }
 
-int _alpm_gpgme_checksig(alpm_handle_t UNUSED *handle, const char UNUSED *path,
-		const char UNUSED *base64_sig, alpm_siglist_t UNUSED *siglist)
+int _alpm_gpgme_checksig(alpm_handle_t *handle, const char UNUSED *path,
+		const char UNUSED *base64_sig, alpm_siglist_t *siglist)
 {
+	siglist->count = 0;
+	handle->pm_errno = ALPM_ERR_MISSING_CAPABILITY_SIGNATURES;
 	return -1;
 }
 #endif /* HAVE_LIBGPGME */
@@ -900,7 +984,7 @@ int _alpm_process_siglist(alpm_handle_t *handle, const char *identifier,
 				_alpm_log(handle, ALPM_LOG_ERROR,
 						_("%s: key \"%s\" is unknown\n"), identifier, name);
 
-				if(_alpm_key_import(handle, result->key.fingerprint) == 0) {
+				if(_alpm_key_import(handle, result->key.uid, result->key.fingerprint) == 0) {
 					retry = 1;
 				}
 
